from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas_fab.backends.exceptions import BackendError


class KinematicsError(BackendError):
    """Base case for exceptions in ``compas_fab.backends.kinematics``."""

    def __init__(self, message):
        super(KinematicsError, self).__init__(message)


class InverseKinematicsError(KinematicsError):
    """Exception raised when no IK solution can be found in Kinematics."""

    def __init__(self):
        message = "No inverse kinematics solution found."
        super(InverseKinematicsError, self).__init__(message)


class CartesianMotionError(KinematicsError):
    """Exception raised when no path can be found."""

    def __init__(self):
        message = "No complete trajectory found."
        super(InverseKinematicsError, self).__init__(message)



from .offset_wrist import inverse_kinematics_offset_wrist
from .offset_wrist import forward_kinematics_offset_wrist

# The following parameters for UR robots are taken from the following website:
# https://www.universal-robots.com/articles/ur/application-installation/dh-parameters-for-calculations-of-kinematics-and-dynamics/

UR10_PARAMS = {'d1': 0.1273,
               'a2': -0.612,
               'a3': -0.5723,
               'd4':  0.163941,
               'd5':  0.1157,
               'd6':  0.0922}

UR10e_PARAMS = {'d1': 0.1807,
                'a2': -0.6127,
                'a3': -0.57155,
                'd4':  0.17415,
                'd5':  0.11985,
                'd6':  0.11655}

UR5_PARAMS = {'d1':  0.089159,
              'a2': -0.42500,
              'a3': -0.39225,
              'd4':  0.10915,
              'd5':  0.09465,
              'd6':  0.0823}

UR5e_PARAMS = {'d1':  0.1625,
               'a2': -0.425,
               'a3': -0.3922,
               'd4':  0.1333,
               'd5':  0.0997,
               'd6':  0.0996}


UR3_PARAMS = {'d1':  0.1519,
              'a2': -0.24365,
              'a3': -0.21325,
              'd4':  0.11235,
              'd5':  0.08535,
              'd6':  0.0819}

UR3e_PARAMS = {'d1': 0.15185,
               'a2': -0.24355,
               'a3': -0.2132,
               'd4':  0.13105,
               'd5':  0.08535,
               'd6':  0.0921}


class OffsetWristKinematics(object):
    """
    """

    def __init__(self, params):
        self.params = params

    def forward(self, joint_values):
        return forward_kinematics_offset_wrist(joint_values, self.params)

    def inverse(self, frame_rcf):
        return inverse_kinematics_offset_wrist(frame_rcf, self.params)


class UR3Kinematics(OffsetWristKinematics):
    def __init__(self):
        params = [UR3_PARAMS[k] for k in ['d1', 'a2', 'a3', 'd4', 'd5', 'd6']]
        super(UR3Kinematics, self).__init__(params)


class UR3eKinematics(OffsetWristKinematics):
    def __init__(self):
        params = [UR3_PARAMS[k] for k in ['d1', 'a2', 'a3', 'd4', 'd5', 'd6']]
        super(UR3eKinematics, self).__init__(params)


class UR5Kinematics(OffsetWristKinematics):
    def __init__(self):
        params = [UR5_PARAMS[k] for k in ['d1', 'a2', 'a3', 'd4', 'd5', 'd6']]
        super(UR5Kinematics, self).__init__(params)


class UR5eKinematics(OffsetWristKinematics):
    def __init__(self):
        params = [UR5e_PARAMS[k] for k in ['d1', 'a2', 'a3', 'd4', 'd5', 'd6']]
        super(UR5eKinematics, self).__init__(params)


class UR10Kinematics(OffsetWristKinematics):
    def __init__(self):
        params = [UR10_PARAMS[k] for k in ['d1', 'a2', 'a3', 'd4', 'd5', 'd6']]
        super(UR10Kinematics, self).__init__(params)


class UR10eKinematics(OffsetWristKinematics):
    def __init__(self):
        params = [UR10_PARAMS[k] for k in ['d1', 'a2', 'a3', 'd4', 'd5', 'd6']]
        super(UR10eKinematics, self).__init__(params)



"""
*******************************************************************************
compas_fab.backends.kinematics
*******************************************************************************

.. module:: compas_fab.backends.kinematics

.. autosummary::
    :toctree: generated/

    AnalyticalInverseKinematics
    AnalyticalPlanCartesianMotion
    InverseKinematicsError
    CartesianMotionError

"""

from __future__ import absolute_import
from .analytical_inverse_kinematics import AnalyticalInverseKinematics  # noqa: F401
from .analytical_plan_cartesian_motion import AnalyticalPlanCartesianMotion  # noqa: F401
from .exceptions import *                 # noqa: F401,F403

__all__ = [name for name in dir() if not name.startswith('_')]



"""
Subscribe to a ROS topic.

COMPAS FAB v0.21.0
"""
import time

import Grasshopper.Kernel
from ghpythonlib.componentbase import executingcomponent as component
from roslibpy import Topic
from scriptcontext import sticky as st

from compas_fab.backends.ros import ROSmsg
from compas_fab.ghpython.components import create_id


class ROSTopicSubscribe(component):
    def RunScript(self, ros_client, topic_name, topic_type, interval, start, stop):
        if not topic_name:
            raise ValueError('Please specify the name of the topic')
        if not topic_type:
            raise ValueError('Please specify the type of the topic')

        if not hasattr(self, 'message_count'):
            self.message_count = 0

        self.interval = interval or 25  # in milliseconds
        self.is_updating = False
        self.is_subscribed = False

        self.msg_key = create_id(self, 'last_msg')
        key = create_id(self, 'topic')

        last_msg = st.get(self.msg_key, None)
        topic = st.get(key, None)

        if ros_client and ros_client.is_connected:
            if start:
                self._unsubscribe(topic)

                topic = Topic(ros_client, topic_name, topic_type)
                topic.subscribe(self.topic_callback)
                time.sleep(0.2)

                st[key] = topic

            if stop:
                self._unsubscribe(topic)

        self.is_subscribed = topic and topic.is_subscribed

        if last_msg:
            last_msg = ROSmsg.parse(last_msg, topic_type)

        if self.is_subscribed:
            self.Message = 'Subscribed, {} messages'.format(self.message_count)
        else:
            self.Message = 'Not subscribed'

        return (topic, last_msg, self.is_subscribed)

    def _unsubscribe(self, topic):
        if topic and topic.is_subscribed:
            topic.unsubscribe()
            time.sleep(0.2)

    def topic_callback(self, msg):
        self.message_count += 1
        st[self.msg_key] = msg

        if self.is_subscribed:
            ghdoc = ghenv.Component.OnPingDocument()     # noqa: F821 This is defined by Grasshopper
            if not self.is_updating and ghdoc.SolutionState != Grasshopper.Kernel.GH_ProcessStep.Process:
                self.is_updating = True
                ghdoc.ScheduleSolution(
                    self.interval, Grasshopper.Kernel.GH_Document.GH_ScheduleDelegate(self.expire_callback))

    def expire_callback(self, ghdoc):
        if ghdoc.SolutionState != Grasshopper.Kernel.GH_ProcessStep.Process:
            ghenv.Component.ExpireSolution(False)        # noqa: F821 This is defined by Grasshopper
        self.is_updating = False



import math
from compas.geometry import argmin
from compas_fab.robots import JointTrajectory
from compas_fab.robots import JointTrajectoryPoint
from compas_fab.backends.interfaces import PlanCartesianMotion
from compas_fab.backends.kinematics.utils import smallest_joint_angles
from compas_fab.backends.kinematics.exceptions import CartesianMotionError


class AnalyticalPlanCartesianMotion(PlanCartesianMotion):
    """
    """

    def __init__(self, client=None):
        self.client = client

    def plan_cartesian_motion(self, robot, frames_WCF, start_configuration=None, group=None, options=None):
        """Calculates a cartesian motion path (linear in tool space).

        Parameters
        ----------
        robot : :class:`compas_fab.robots.Robot`
            The robot instance for which the cartesian motion path is being calculated.
        frames_WCF : list of :class:`compas.geometry.Frame`
            The frames through which the path is defined.
        start_configuration : :class:`Configuration`, optional
            The robot's full configuration, i.e. values for all configurable
            joints of the entire robot, at the starting position.
        group : str, optional
            The planning group used for calculation.
        options : dict, optional
            Dictionary containing kwargs for arguments specific to
            the client being queried.

        Returns
        -------
        :class:`compas_fab.robots.JointTrajectory`
            The calculated trajectory.

        Notes
        -----
        This will only work with robots that have 6 revolute joints.
        """
        # what is the expected behaviour of that function?
        # - Return all possible paths or select only the one that is closest to the start_configuration?
        # - Do we use a stepsize to sample in between frames or use only the input frames?

        # convert the frame WCF to RCF
        base_frame = robot.get_base_frame(group=group, full_configuration=start_configuration)
        frames_RCF = [base_frame.to_local_coordinates(frame_WCF) for frame_WCF in frames_WCF]

        options = options or {}
        options.update({"keep_order": True})

        configurations_along_path = []
        for frame in frames_RCF:
            configurations = list(robot.iter_inverse_kinematics(frame, options=options))
            configurations_along_path.append(configurations)

        paths = []
        for configurations in zip(*configurations_along_path):
            if all(configurations):
                paths.append(configurations)

        if not len(paths):
            raise CartesianMotionError("No complete trajectory found.")

        # now select the path that is closest to the start configuration.
        first_configurations = [path[0] for path in paths]
        diffs = [sum([abs(d) for d in start_configuration.iter_differences(c)]) for c in first_configurations]
        idx = argmin(diffs)

        path = paths[idx]
        path = self.smooth_configurations(path)
        trajectory = JointTrajectory()
        trajectory.fraction = len(path)/len(frames_RCF)
        trajectory.joint_names = path[0].joint_names
        trajectory.points = [JointTrajectoryPoint(config.joint_values, config.joint_types) for config in path]
        trajectory.start_configuration = robot.merge_group_with_full_configuration(path[0], start_configuration, group)
        return trajectory

    def smooth_configurations(self, configurations):
        joint_values_corrected = []
        prev = smallest_joint_angles(configurations[0].joint_values)
        joint_values_corrected.append(prev)

        for i in range(1, len(configurations)):
            curr = configurations[i].joint_values
            corrected = []
            for p, c in zip(prev, curr):
                c1 = c/abs(c) * (abs(c) % (2 * math.pi))
                c2 = c1 - 2 * math.pi
                c3 = c1 + 2 * math.pi
                values = [c1, c2, c3]
                diffs = [math.fabs(p - v) for v in values]
                idx = argmin(diffs)
                corrected.append(values[idx])
            prev = corrected
            joint_values_corrected.append(corrected)

        # now that the values are continuous, try to bring them all "down"
        for i, j in enumerate(zip(*joint_values_corrected)):
            minj, maxj = min(j), max(j)
            v1 = minj + maxj
            v2 = minj - 2 * math.pi + maxj - 2 * math.pi
            v3 = minj + 2 * math.pi + maxj + 2 * math.pi
            values = [math.fabs(v) for v in [v1, v2, v3]]
            idx = argmin(values)
            if idx == 1:
                for k in range(len(joint_values_corrected)):
                    joint_values_corrected[k][i] -= 2 * math.pi
            elif idx == 2:
                for k in range(len(joint_values_corrected)):
                    joint_values_corrected[k][i] += 2 * math.pi

        for i in range(len(joint_values_corrected)):
            configurations[i].joint_values = joint_values_corrected[i]
        return configurations



"""
********************************************************************************
compas_fab.backends
********************************************************************************

.. currentmodule:: compas_fab.backends

This package contains classes backends for simulation, planning and execution.

V-REP
-----

.. autosummary::
    :toctree: generated/
    :nosignatures:

    VrepClient
    VrepPlanner

ROS
---

.. autosummary::
    :toctree: generated/
    :nosignatures:

    RosClient
    RosFileServerLoader
    MoveItPlanner

PyBullet
--------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    PyBulletClient
    PyBulletPlanner
    PyBulletError


Analytical Kinematics
---------------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    AnalyticalPyBulletClient
    AnalyticalInverseKinematics
    AnalyticalPlanCartesianMotion


Long-running tasks
------------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    FutureResult
    CancellableFutureResult

Exceptions
----------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    BackendError
    CollisionError
    InverseKinematicsError
    RosError
    RosValidationError
    VrepError


Interfaces
----------

For details about integrating new backends, check
the :ref:`architecture` documentation.

"""

import compas

from .exceptions import (
    BackendError,
)
from .tasks import (
    FutureResult,
    CancellableFutureResult,
)
from .ros.client import (
    RosClient,
)
from .ros.exceptions import (
    RosError,
    RosValidationError,
)
from .ros.fileserver_loader import (
    RosFileServerLoader,
)
from .ros.planner import (
    MoveItPlanner,
)
from .vrep.client import (
    VrepClient,
)
from .vrep.helpers import (
    VrepError,
)
from .vrep.planner import (
    VrepPlanner,
)

if not compas.IPY:
    from .pybullet.client import (
        PyBulletClient,
    )
    from .pybullet.exceptions import (
        CollisionError,
        InverseKinematicsError,
        PyBulletError,
    )
    from .pybullet.planner import (
        PyBulletPlanner,
    )
    from .kinematics.client import (
        AnalyticalPyBulletClient,
    )
    from .kinematics import (
        AnalyticalInverseKinematics,
    )
    from .kinematics.client import (
        AnalyticalPlanCartesianMotion,
    )

__all__ = [
      'BackendError',
      'CancellableFutureResult',
      'FutureResult',
      'MoveItPlanner',
      'RosClient',
      'RosError',
      'RosFileServerLoader',
      'RosValidationError',
      'VrepClient',
      'VrepError',
      'VrepPlanner',
    ]

if not compas.IPY:
    __all__ += [
        'CollisionError',
        'InverseKinematicsError',
        'PyBulletClient',
        'PyBulletError',
        'PyBulletPlanner',
        'AnalyticalPyBulletClient',
        'AnalyticalInverseKinematics',
        'AnalyticalPlanCartesianMotion',

    ]



"""
********************************************************************************
compas_fab.utilities
********************************************************************************

.. currentmodule:: compas_fab.utilities

Package containing a set of utility functions.

File system functions
=====================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    list_files_in_directory
    read_csv_to_dictionary
    read_data_from_json
    read_data_from_pickle
    write_data_to_json
    write_data_to_pickle

Numerical functions
===================

.. autosummary::
    :toctree: generated/
    :nosignatures:

    allclose
    arange
    clamp
    diffs
    map_range
    range_geometric_row

Other functions
===============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    argmax
    argmin
    argsort
    LazyLoader
    sign

"""

from .file_io import (
    read_csv_to_dictionary,
    read_data_from_json,
    read_data_from_pickle,
    write_data_to_json,
    write_data_to_pickle,
)
from .filesystem import (
    list_files_in_directory,
)
from .lazy_loader import (
    LazyLoader,
)
from .numbers import (
    allclose,
    arange,
    argmax,
    argmin,
    clamp,
    diffs,
    map_range,
    range_geometric_row,
)
from .utilities import (
    argsort,
    sign,
)

__all__ = [
    'LazyLoader',
    'allclose',
    'arange',
    'argmax',
    'argmin',
    'argsort',
    'clamp',
    'diffs',
    'list_files_in_directory',
    'map_range',
    'range_geometric_row',
    'read_csv_to_dictionary',
    'read_data_from_json',
    'read_data_from_pickle',
    'sign',
    'write_data_to_json',
    'write_data_to_pickle',
]
